import pytest
from promptify import Prompter
from promptify import OpenAI
from typing import List, Optional, Union, Dict
import os


class TestPrompter:
    @pytest.fixture
    def model(self):
        model = OpenAI(api_key="", api_wait=1, api_retry=1)
        return model

    def test_custom_template(self, model):

        # replace the template path with own path, this is just for testing
        prompter = Prompter(
            model=model, template="/Users/stoicbatman/Desktop/pytest_project/ner.jinja"
        )
        output = prompter.fit(
            "Elon Reeve Musk FRS is a business. He is the founder of SpaceX; Tesla, Inc.; Twitter, Inc.; Neuralink and OpenAI",
            domain="general",
            labels=None,
        )
        assert isinstance(output, list)
        assert isinstance(output[0]["parsed"], Dict)
        assert isinstance(output[0]["parsed"]["data"]["completion"][0]["T"], str)
        assert isinstance(output[0]["parsed"]["data"]["completion"][0]["E"], str)

    def test_generate_prompt(self, model):
        prompter = Prompter(model=model, template="ner.jinja")
        prompt = prompter.generate_prompt(
            "Elon Reeve Musk FRS is a business. He is the founder of SpaceX; Tesla, Inc.; Twitter, Inc.; Neuralink and OpenAI",
            domain="general",
            labels=None,
        )
        assert isinstance(prompt, str)

    def test_fit(self, model):
        prompter = Prompter(model=model, template="ner.jinja")
        output = prompter.fit(
            "Elon Reeve Musk FRS is a business. He is the founder of SpaceX; Tesla, Inc.; Twitter, Inc.; Neuralink and OpenAI",
            domain="general",
            labels=None,
        )
        assert isinstance(output, list)
        assert isinstance(output[0]["parsed"], Dict)
        assert isinstance(output[0]["parsed"]["data"]["completion"][0]["T"], str)
        assert isinstance(output[0]["parsed"]["data"]["completion"][0]["E"], str)

    def test_raw_fit(self, model):
        prompter = Prompter(model=model, raw_prompt=True)
        output = prompter.fit("quick brown fox jump over")
        assert isinstance(output, list)
        assert isinstance(output[0]["text"], str)

    def test_load_template(self, model):
        prompter = Prompter(model=model)
        template_data = prompter.load_template("ner.jinja")
        assert "template_name" in template_data
        assert "template_dir" in template_data
        assert "environment" in template_data
        assert "template" in template_data

    def test_get_available_templates(self, model):
        prompter = Prompter(model)
        templates_path = os.path.join(
            os.path.dirname(os.path.realpath(".")), "codes", "templates"
        )
        templates = prompter.get_available_templates(templates_path)
        assert isinstance(templates, dict)
        for key, value in templates.items():
            assert key.endswith(".jinja")
            assert value.endswith(".jinja")

    def test_list_templates(self, model):
        prompter = Prompter(model=model, template="ner.jinja")
        loader = prompter.load_template("ner.jinja")
        templates = loader["environment"].list_templates()
        assert isinstance(templates, list)
        assert len(templates) > 0

    def test_template_variables(self, model):
        prompter = Prompter(model=model, template="ner.jinja")
        loader = prompter.load_template("ner.jinja")
        variables = prompter.get_template_variables(
            loader["environment"], loader["template_name"]
        )
        assert isinstance(variables, set)
        assert len(variables) > 0

    def test_update_default_variable_values(self, model):
        prompter = Prompter(model=model, template="ner.jinja")
        new_defaults = {"description": "test description", "domain": "test domain"}
        prompter.update_default_variable_values(new_defaults)
        assert prompter.default_variable_values == new_defaults

    def test_missing_template_path_error(self, model):
        with pytest.raises(ValueError):
            prompter = Prompter(model=model)
            prompter.load_template("non_existent_template.jinja")



import pytest
from promptify import OpenAI
from typing import List, Dict
from promptify import Parser

@pytest.fixture
def openai_complete():
    openai_complete = OpenAI(api_key="", api_wait=5, api_retry=5)
    return openai_complete


def test_supported_models(openai_complete):
    models = openai_complete.supported_models()
    assert "text-davinci-003" in models
    assert "text-curie-001" in models
    assert "text-babbage-001" in models
    assert "text-ada-001" in models


def test_invalid_model(openai_complete):
    with pytest.raises(ValueError):
        openai_complete.set_model("invalid-model")


def test_get_description(openai_complete):
    description = openai_complete.get_description()
    assert isinstance(description, str)


def test_get_endpoint(openai_complete):
    endpoint = openai_complete.get_endpoint()
    assert isinstance(endpoint, str)


def test_run(openai_complete):
    prompts = ["Hello, my name is", "The quick brown fox jumps over the"]
    results = openai_complete.run(prompts)
    assert len(results) == 2
    for result in results:
        assert "text" in result["choices"][0]
        assert result["choices"][0]["text"] != ""
        assert isinstance(result["choices"][0]["text"], str)


def test_execute_with_retry(openai_complete):
    prompts = [
        """You are a highly intelligent and accurate medical domain Named-entity recognition(NER) system. You take Passage as input and your task is to recognize and extract specific types of medical domain named entities in that given passage and classify into a set of entity types. Your output valid json and format is only [{{'T': type of entity from predefined entity types, 'E': entity in the input text}},...,{{'branch' : Appropriate branch of the passage ,'group': Appropriate Group of the passage}}] form, no other form.
                Examples: Input: The patient had abdominal pain and 30-pound weight loss then developed jaundice. He had epigastric pain. A thin-slice CT scan was performed, which revealed a pancreatic mass with involved lymph nodes and ring enhancing lesions with liver metastases Output: [[{'T': 'SYMPTOM', 'E': 'abdominal pain'}, {'T': 'QUANTITY', 'E': '30-pound'}, {'T': 'SYMPTOM', 'E': 'jaundice'}, {'T': 'SYMPTOM', 'E': "epigastric pain"}, {'T': 'TEST', 'E': 'thin-slice CT scan'}, {'T': 'ANATOMY', 'E': "pancreatic mass"}, {'T': 'ANATOMY', 'E': 'ring enhancing lesions'}, {'T': 'ANATOMY', 'E': 'liver'}, {'T': 'DISEASE', 'E': 'metastases'}, {'branch': 'Health', 'group': 'Clinical medicine'}]] Input: Several important diseases of the nervous system are associated with dysfunctions of the dopamine system, and some of the key medications used to treat them work by altering the effects of dopamine. Parkinson's disease, a degenerative condition causing tremor and motor impairment, is caused by a loss of dopamine-secreting neurons in an area of the midbrain called the substantia nigra. Its metabolic precursor L-DOPA can be manufactured; Levodopa, a pure form of L-DOPA, is the most widely used treatment for Parkinson's. Output:"""
    ]
    results = openai_complete.execute_with_retry(prompts)
    for result in results:
        result_s = openai_complete.model_output(result, 20)["parsed"][
            "data"
        ]["completion"][0]

        assert isinstance(result_s, List)
        assert isinstance(result_s[0], Dict)


def test_model_output_with_parser(openai_complete):
    prompts = [
        """You are a highly intelligent and accurate medical domain Named-entity recognition(NER) system. You take Passage as input and your task is to recognize and extract specific types of medical domain named entities in that given passage and classify into a set of entity types. Your output valid json and format is only [{{'T': type of entity from predefined entity types, 'E': entity in the input text}},...,{{'branch' : Appropriate branch of the passage ,'group': Appropriate Group of the passage}}] form, no other form.
                Examples: Input: The patient had abdominal pain and 30-pound weight loss then developed jaundice. He had epigastric pain. A thin-slice CT scan was performed, which revealed a pancreatic mass with involved lymph nodes and ring enhancing lesions with liver metastases Output: [[{'T': 'SYMPTOM', 'E': 'abdominal pain'}, {'T': 'QUANTITY', 'E': '30-pound'}, {'T': 'SYMPTOM', 'E': 'jaundice'}, {'T': 'SYMPTOM', 'E': "epigastric pain"}, {'T': 'TEST', 'E': 'thin-slice CT scan'}, {'T': 'ANATOMY', 'E': "pancreatic mass"}, {'T': 'ANATOMY', 'E': 'ring enhancing lesions'}, {'T': 'ANATOMY', 'E': 'liver'}, {'T': 'DISEASE', 'E': 'metastases'}, {'branch': 'Health', 'group': 'Clinical medicine'}]] Input: Several important diseases of the nervous system are associated with dysfunctions of the dopamine system, and some of the key medications used to treat them work by altering the effects of dopamine. Parkinson's disease, a degenerative condition causing tremor and motor impairment, is caused by a loss of dopamine-secreting neurons in an area of the midbrain called the substantia nigra. Its metabolic precursor L-DOPA can be manufactured; Levodopa, a pure form of L-DOPA, is the most widely used treatment for Parkinson's. Output:"""
    ]

    results = openai_complete.run(prompts)
    for result in results:
        result_s = openai_complete.model_output(result, 20)["parsed"][
            "data"
        ]["completion"][0]

        assert isinstance(result_s, List)
        assert isinstance(result_s[0], Dict)


def test_single_double_quote(openai_complete):
    prompts = [
        """You are a highly intelligent and accurate medical domain Named-entity recognition(NER) system. You take Passage as input and your task is to recognize and extract specific types of medical domain named entities in that given passage and classify into a set of entity types. Your output valid json and format is only [{{'T': type of entity from predefined entity types, 'E': entity in the input text}},...,{{'branch' : Appropriate branch of the passage ,'group': Appropriate Group of the passage}}] form, no other form.
                Examples: Input: The patient had abdominal pain and 30-pound weight loss then developed jaundice. He had epigastric pain. A thin-slice CT scan was performed, which revealed a pancreatic mass with involved lymph nodes and ring enhancing lesions with liver metastases Output: [[{'T': 'SYMPTOM', 'E': 'abdominal pain'}, {'T': 'QUANTITY', 'E': '30-pound'}, {'T': 'SYMPTOM', 'E': 'jaundice'}, {'T': 'SYMPTOM', 'E': "epigastric pain"}, {'T': 'TEST', 'E': 'thin-slice CT scan'}, {'T': 'ANATOMY', 'E': "pancreatic mass"}, {'T': 'ANATOMY', 'E': 'ring enhancing lesions'}, {'T': 'ANATOMY', 'E': 'liver'}, {'T': 'DISEASE', 'E': 'metastases'}, {'branch': 'Health', 'group': 'Clinical medicine'}]] Input: Several important diseases of the nervous system are associated with dysfunctions of the dopamine system, and some of the key medications used to treat them work by altering the effects of dopamine. Parkinson's disease, a degenerative condition causing tremor and motor impairment, is caused by a loss of dopamine-secreting neurons in an area of the midbrain called the substantia nigra. Its metabolic precursor L-DOPA can be manufactured; Levodopa, a pure form of L-DOPA, is the most widely used treatment for Parkinson's. Output:"""
    ]

    parser = Parser()
    results = openai_complete.run(prompts)
    result_filterd = eval(parser.escaped_(results[0]["choices"][0]["text"]))
    assert isinstance(result_filterd[0], List)
    assert isinstance(result_filterd[0][0], Dict)



from abc import ABCMeta, abstractmethod
from typing import List, Optional, Union, Dict
import tenacity


class Model(metaclass=ABCMeta):

    """
    Abstract base class for a large language model(llm).

    Attributes
    ----------
    name : str
        The name of the language model.
    description : str
        A brief description of the language model.

    Methods
    -------
    __init__(api_key, model, api_wait=None, api_retry=None, **kwargs) -> None:
        Initializes the Model class with the required parameters and verifies the model is supported by the endpoint.
    supported_models() -> List[str]:
        Abstract method to return a list of supported models for the endpoint.
    _verify_model() -> None:
        Abstract method to verify if the model is supported by the endpoint.
    set_key(api_key) -> None:
        Abstract method to set the endpoint API key.
    set_model(model) -> None:
        Abstract method to set the model name for the endpoint.
    get_description() -> str:
        Abstract method to get the model description.
    get_endpoint() -> str:
        Abstract method to get the model endpoint.
    get_parameters() -> Dict[str, str]:
        Abstract method to get the model parameters.
    run(prompts) -> List[str]:
        Abstract method to run the language model on the given list of prompts and return the list of responses.
    model_output(response) -> Any:
        Abstract method to get the model output from the response.
    _retry_decorator() -> tenacity.Retry:
        Decorator function for retrying API requests if they fail.
    execute_with_retry(*args, **kwargs) -> List[str]:
        Decorated version of the `run` method with the retry logic.


    Examples
    --------
    >>> class MyModel(Model):
    ...     def __init__(self, api_key, model, api_wait=None, api_retry=None, **kwargs):
    ...         super().__init__(api_key, model, api_wait, api_retry, **kwargs)
    ...
    ...     @classmethod
    ...     def supported_models(cls) -> List[str]:
    ...         return ['gpt', 'davinci']
    ...
    ...     def _verify_model(self):
    ...         assert self.model in self.supported_models(), f"{self.model} is not a supported model"
    ...
    ...     def set_key(self, api_key: str):
    ...         self.api_key = api_key
    ...
    ...     def set_model(self, model: str):
    ...         self.model = model
    ...
    ...     def get_description(self) -> str:
    ...         return self.description
    ...
    ...     def get_endpoint(self) -> str:
    ...         return "https://api.openai.com/v1/completions"
    ...
    ...     def get_parameters(self) -> Dict[str, str]:
    ...         return {"model": self.model, "prompt": "", "temperature": "0.7"}
    ...
    ...     def run(self, prompts: List[str]) -> List[str]:
    ...         # Send the request to OpenAI's API
    ...         response = requests.post(
    ...             self.get_endpoint(),
    ...             headers={
    ...                 "Content-Type": "application/json",
    ...                 "Authorization": f"Bearer {self.api_key}",
    ...             },
    ...             json=self.get_parameters(),
    ...         )
    ...
    ...         # Get the output from the response
    ...         output = self.model_output(response)
    ...
    ...         # Return the output
    ...         return output
    ...
    ...     def model_output(self, response):
    ...         return response.json()['choices'][0]['text']

    Notes
    -----
    This class is an abstract base class for creating large language model classes.
    It provides common methods and attributes that can be used by different llms
    classes to make calls to llms API more streamlined.

    """

    name = ""
    description = ""

    def __init__(
        self,
        api_key: str,
        model: str,
        api_wait: int = 60,
        api_retry: int = 6,
        **kwargs
    ):
        """
        Initializes the Model class with the required parameters and verifies the model is supported by the endpoint.

        Parameters
        ----------
        api_key : str
            The API key if needed for the endpoint.
        model : str
            The name of the LLM model to use for the endpoint.
        api_wait : int, optional
            Maximum wait time for an API request before retrying (in seconds), by default 60.
        api_retry : int, optional
            Number of times to retry an API request before failing, by default 6.
        **kwargs : dict
            Additional arguments to be passed to the Model API call.

        Notes
        -----
        This method initializes the Model class with the required parameters and verifies that the given model is supported by the endpoint. It sets the values of `api_key`, `model`, `api_wait`, and `api_retry` attributes of the class.

        Examples
        --------
        >>> my_model = MyModel(api_key="my_api_key", model="davinci")

        """

        self.api_key = api_key
        self.model = model
        self.api_wait = api_wait
        self.api_retry = api_retry
        self._verify_model()
        self.set_key(api_key)

    @abstractmethod
    def supported_models(self):
        """
        Get a list of supported models for the endpoint.

        Returns
        -------
        List[str]
            A list of supported models for the endpoint.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def supported_models(self):
        ...         return ['gpt', 'davinci']
        """

        raise NotImplementedError

    @abstractmethod
    def _verify_model(self):
        """
        Verify the model is supported by the endpoint.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def _verify_model(self):
        ...         assert self.model in self.supported_models(), f"{self.model} is not a supported model"
        """

        raise NotImplementedError

    @abstractmethod
    def set_key(self, api_key: str):
        """
        Set endpoint API key if needed.

        Parameters
        ----------
        api_key : str
            The API key to set.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def set_key(self, api_key: str):
        ...         self.api_key = api_key
        """

        raise NotImplementedError

    @abstractmethod
    def set_model(self, model: str):
        """
        Set model name for the endpoint.

        Parameters
        ----------
        model : str
            The model name to set.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def set_model(self, model: str):
        ...         self.model = model
        """

        raise NotImplementedError

    @abstractmethod
    def get_description(self) -> str:
        """
        Get model description.

        Returns
        -------
        str
            A string containing the model description.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def get_description(self) -> str:
        ...         return self.description
        """

        raise NotImplementedError

    @abstractmethod
    def get_endpoint(self) -> str:
        """
        Get model endpoint.

        Returns
        -------
        str
            A string containing the model endpoint.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def get_endpoint(self) -> str:
        ...         return "https://api.openai.com/v1/completions"
        """

        raise NotImplementedError

    @abstractmethod
    def get_parameters(self) -> Dict[str, str]:
        """
        Get model parameters.

        Returns
        -------
        Dict[str, str]
            A dictionary containing the model parameters.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def get_parameters(self) -> Dict[str, str]:
        ...         return {"model": self.model, "prompt": "", "temperature": "0.7"}
        """

        raise NotImplementedError
    
    @abstractmethod
    def run(self, prompts: List[str]) -> List[str]:
        """
        Run the LLM on the given prompt list.

        Parameters
        ----------
        prompts : List[str]
            A list of prompts to run on the LLM.

        Returns
        -------
        List[str]
            A list of responses from the LLM.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def run(self, prompts: List[str]) -> List[str]:
        ...         # Send the request to OpenAI's API
        ...         response = requests.post(
        ...             self.get_endpoint(),
        ...             headers={
        ...                 "Content-Type": "application/json",
        ...                 "Authorization": f"Bearer {self.api_key}",
        ...             },
        ...             json=self.get_parameters(),
        ...         )
        ...
        ...         # Get the output from the response
        ...         output = self.model_output(response)
        ...
        ...         # Return the output
        ...         return output
        """

        raise NotImplementedError

    @abstractmethod
    def model_output(self, response):
        """
        Get the model output from the response.

        Parameters
        ----------
        response : requests.Response
            The response from the API call.

        Notes
        -----
        This method is an abstract method and must be implemented in the derived classes.

        Examples
        --------
        >>> class MyModel(Model):
        ...     def model_output(self, response):
        ...         return response.json()['choices'][0]['text']
        """
        raise NotImplementedError

    def _retry_decorator(self):
        """
        Decorator function for retrying API requests if they fail.

        Returns
        -------
        tenacity.Retrying
            A decorator function for retrying API requests.

        Notes
        -----
        This method is a decorator function for retrying API requests using tenacity.
        """

        return tenacity.retry(
            wait=tenacity.wait_random_exponential(
                multiplier=0.3, exp_base=3, max=self.api_wait
            ),
            stop=tenacity.stop_after_attempt(self.api_retry),
        )

    def execute_with_retry(self, *args, **kwargs):
        """

        Decorated version of the run method with the retry logic.

        Parameters
        ----------
        *args : tuple
            A tuple of arguments to pass to the `run` method.
        **kwargs : dict
            A dictionary of keyword arguments to pass to the `run` method.

        Returns
        -------
        Any
            The output of the `run` method.

        Notes
        -----
        This method is a decorated version of the `run` method with the retry logic.
        """

        decorated_run = self._retry_decorator()(self.run)
        return decorated_run(*args, **kwargs)



from typing import List, Dict
from promptify.models.text2text.api.base_model import Model


class MockModel(Model):

    """
    A mock model for testing purposes.

    This is a base class for models that provides a common interface for interacting with models.
    It defines methods for setting model parameters, running the model, and retrieving model output.

    Attributes
    ----------
    name : str
        The name of the model.
    description : str
        A description of the model.

    Methods
    -------
    supported_models() -> List[str]:
        Returns a list of supported models.

    set_key(api_key: str):
        Sets the API key for the model.

    set_model(model: str):
        Sets the model to use.

    get_description() -> str:
        Returns the description of the model.

    get_endpoint() -> str:
        Returns the endpoint for the model.

    get_parameters() -> Dict[str, str]:
        Returns the parameters for the model.

    run(prompts: List[str]) -> List[str]:
        Runs the model on the given prompts and returns a list of responses.

    model_output(response):
        Processes the model's response and returns it as output.

    """

    name = "mock_model"
    description = "Mock model for testing purposes"

    
    def supported_models(self) -> List[str]:
        """
        Returns a list of supported models.

        Returns
        -------
        List[str]
            A list of supported models.
        """
        return ['mock_model']

    def _verify_model(self):
        """
        Checks if the model is supported.

        Raises
        ------
        ValueError
            If the model is not supported.

        """
        if self.model not in self.supported_models():
            raise ValueError(f"Unsupported model: {self.model}")

    def set_key(self, api_key: str):
        """
        Sets the API key for the model.

        Parameters
        ----------
        api_key : str
            The API key for the model.

        """
        self.api_key = api_key

    def set_model(self, model: str):
        """
        Sets the model to use.

        Parameters
        ----------
        model : str
            The name of the model to use.

        Raises
        ------
        ValueError
            If the model is not supported.

        """

        self.model = model
        self._verify_model()

    def get_description(self) -> str:
        """
        Returns the description of the model.

        Returns
        -------
        str
            The description of the model.

        """
        return self.description

    def get_endpoint(self) -> str:
        """
        Returns the endpoint for the model.

        Returns
        -------
        str
            The endpoint for the model.

        """
        return "https://mock.endpoint/"

    def get_parameters(self) -> Dict[str, str]:
        """
        Returns the parameters for the model.

        Returns
        -------
        Dict[str, str]
            A dictionary of the model's parameters.

        """
        return {"param": "value"}

    def run(self, prompts: List[str]) -> List[str]:
        """
        Runs the model on the given prompts and returns a list of responses.

        Parameters
        ----------
        prompts : List[str]
            A list of prompts to run the model on.

        Returns
        -------
        List[str]
            A list of responses from the model.

        """
        return {'text': 'response', 'parsed': {'data': {"completion" : []}}}

    def model_output(self, response, max_completion_length = None):
        """
        Processes the model's response and returns it as output.

        Parameters
        ----------
        response :
            The response from the model.

        Returns
        -------
        Output of the model.

        """

        response = {'text': 'response', 'parsed': {'data': {"completion" : []}}}


        
        return response

    def model_output_raw(self, response, max_completion_length = None):
        """
        Processes the model's response and returns it as output.

        Parameters
        ----------
        response :
            The response from the model.

        Returns
        -------
        Output of the model.

        """

        response = {'text': 'response', 'parsed': {'data': {"completion" : []}}}
        return response



from typing import Dict, List, Optional, Tuple, Union
import openai
import json
import tiktoken
from promptify.parser.parser import Parser
from promptify.models.text2text.api.base_model import Model


class OpenAI(Model):
    name = "OpenAI"
    description = "OpenAI API for text completion using various models"

    SUPPORTED_MODELS = {
        "completion_models": set(
            [
                "text-davinci-003",
                "davinci",
                "text-davinci-001",
                "ada",
                "text-curie-001",
                "text-ada-001",
                "text-babbage-001",
                "curie",
                "text-davinci-002",
            ]
        ),
        "chat_models": set(
            [
                "gpt-4-0314",
                "gpt-3.5-turbo-16k-0613",
                "gpt-3.5-turbo-0301",
                "gpt-3.5-turbo-16k",
                "gpt-4",
                "gpt-3.5-turbo",
                "gpt-3.5-turbo-0613",
                "gpt-4-0613",
            ]
        ),
    }

    def __init__(
        self,
        api_key: str,
        model: str = "gpt-3.5-turbo",
        temperature: float = 0.7,
        top_p: float = 1,
        n: int = 1,
        stop: Optional[Union[str, List[str]]] = None,
        presence_penalty: float = 0,
        frequency_penalty: float = 0,
        logit_bias: Optional[Dict[str, int]] = None,
        request_timeout: Union[float, Tuple[float, float]] = None,
        api_wait=60,
        api_retry=6,
        json_depth_limit: int = 20,
    ):
        super().__init__(api_key, model, api_wait, api_retry)

        self.temperature = temperature
        self.top_p = top_p
        self.n = n
        self.stop = stop
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.logit_bias = logit_bias or {}
        self.request_timeout = request_timeout
        self.json_depth_limit = json_depth_limit
        self.set_key(api_key)
        self._verify_model()
        self._initialize_encoder()
        self._initialize_parser()
        self.parameters = self.get_parameters()

    def set_key(self, api_key: str):
        self._openai = openai
        self._openai.api_key = api_key

    def _verify_model(self):
        model_type = (
            "completion_models"
            if self.model in self.SUPPORTED_MODELS["completion_models"]
            else "chat_models"
        )
        if self.model not in self.SUPPORTED_MODELS[model_type]:
            raise ValueError(f"Unsupported model: {self.model}")
        self.model_type = model_type

    def _initialize_encoder(self):
        self.encoder = tiktoken.encoding_for_model(self.model)

    def _initialize_parser(self):
        self.parser = Parser()

    def set_model(self, model: str):
        self.model = model
        self._verify_model()

    def supported_models(self):
        return list(itertools.chain(*self.SUPPORTED_MODELS.values()))

    def get_parameters(self):
        return {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "n": self.n,
            "stop": self.stop,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
            "logit_bias": self.logit_bias,
            "request_timeout": self.request_timeout,
        }

    def get_description(self):
        return self.description

    def get_endpoint(self):
        model = self._openai.Model.retrieve(self.model)
        return model["id"]

    def run(self, prompt: str):
        if self.model_type == "chat_models":
            return self._chat_api(prompt)
        elif self.model_type == "completion_models":
            return self._completion_api(prompt)


    def _completion_api(self, prompt: str):
        self.parameters["prompt"] = prompt
        self.parameters["max_tokens"] = self._calculate_max_tokens(prompt)
        response = self._openai.Completion.create(
            model=self.model,
            **self.parameters,
        )
        return response

    def _chat_api(self, prompt: str):
        prompt_template = [
            {"role": "system", "content": "you are a helpful assistant."},
            {"role": "user", "content": prompt},
        ]

        self.parameters["max_tokens"] = self._calculate_max_tokens(prompt_template)
        self.parameters["messages"] = prompt_template
        response = self._openai.ChatCompletion.create(
            model=self.model,
            **self.parameters,
        )
        return response

    def _calculate_max_tokens(self, prompt: str) -> int:
        prompt_tokens = len(self.encoder.encode(str(prompt)))
        max_tokens = self._default_max_tokens(self.model) - prompt_tokens
        return max_tokens

    def _default_max_tokens(self, model_name: str) -> int:
        token_dict = {
            "text-babbage-001": 2040,
            "text-ada-001": 2048,
            "ada": 2048,
            "babbage": 2048,
            "text-curie-001": 2048,
            "curie": 2048,
            "davinci": 2048,
            "code-cushman-002": 2048,
            "code-cushman-001": 2048,
            "text-davinci-003": 4000,
            "gpt-3.5-turbo": 4096,
            "gpt-3.5-turbo-0301": 4096,
            "gpt-3.5-turbo-0613": 4096,
            "text-davinci-002": 4096,
            "code-davinci-002": 8000,
            "code-davinci-001": 8000,
            "gpt-4": 8192,
            "gpt-4-0314": 8192,
            "gpt-4-0613": 8192,
            "gpt-3.5-turbo-16k": 16385,
            "gpt-3.5-turbo-16k-0613": 16385,
            "gpt-4-32k": 32768,
            "gpt-4-32k-0314": 32768,
            "gpt-4-32k-0613": 32768,
        }
        return token_dict[model_name]

    def model_output_raw(self, response: Dict) -> Dict:
        data = {}
        if self.model_type == "chat_models":
            data["text"] = response["choices"][0]["message"]["content"].strip(" \n")
        elif self.model_type == "completion_models":
            data["text"] = response["choices"][0]["text"]

        data["usage"] = dict(response["usage"])
        return data

    def model_output(self, response, json_depth_limit: int) -> Dict:
        data = self.model_output_raw(response)
        data["parsed"] = self.parser.fit(data["text"], json_depth_limit)

        return data



import itertools
from operator import itemgetter
from typing import Any, Dict, List, Union, Optional
import re
import ast


class Parser:
    """
    A class to parse incomplete JSON objects and provide possible completions.

    Methods
    -------
    is_valid_json() -> bool:
        Checks if a string is valid JSON.

    get_combinations() -> List[str]:
        Returns all possible combinations of } and ] characters up to length n.

    complete_json_object() -> Any:
        Completes a JSON object string by appending a completion string.

    get_possible_completions() -> Union[Dict[str, Any], List[Any]]:
        Returns a list of possible completions for a JSON object string.

    fit() -> Dict[str, Any]:
        Tries to parse the input JSON string and complete it if it is incomplete.

    find_max_length() -> Dict[str, List[Any]]:
        Returns a dictionary containing the element with the maximum length in the input list.
    """

    def __init__(self):
        pass

    def is_valid_json(self, input_str: str) -> bool:
        """
        Check if the input string is valid JSON.

        Parameters
        ----------
        input_str : str
            The string to check for validity.

        Returns
        -------
        bool
            Returns True if the input string is valid JSON, otherwise False.

        Notes
        -----
        This function uses the `json` module to check if the input string is valid JSON.
        It evaluates the input string using `eval()`, and if it successfully loads
        a JSON object (either a dictionary or a list), it returns True. Otherwise, it
        returns False.

        Examples
        --------
        >>> validator = Parser()
        >>> validator.is_valid_json('{"name": "Alice", "age": 30}')
        True
        >>> validator.is_valid_json('[1, 2, 3, 4]')
        True
        >>> validator.is_valid_json('{"name": "Bob", "age": }')
        False
        >>> validator.is_valid_json('not a JSON string')
        False
        """
        try:
            output = eval(input_str)
            if isinstance(output, (dict, list)):
                return True
            else:
                return False
        except Exception:
            return False

    def escaped_(self, data: str) -> str:
        if "'" in data:
            escaped_str = re.sub(r"(?<=\w)(')(?=\w)", r"\"", data)
        else:
            escaped_str = re.sub(r'(?<=\w)(")(?=\w)', r"\'", data)

        return escaped_str

    def get_combinations(
        self, candidate_marks: List[str], n: int, should_end_mark: Optional[str] = None
    ) -> List[str]:
        """
        Return all possible combinations of candidate marks up to length n.

        Parameters
        ----------
        candidate_marks : list of str
            Candidate marks to combine.
        n : int
            The maximum length of the combinations.
        should_end_mark : str or None, optional
            If provided, only combinations that end with this mark will be returned, by default None.

        Returns
        -------
        list of str
            A list of all possible combinations of candidate marks up to length n.
        """
        combinations = []
        for i in range(1, n):
            for comb in itertools.product(candidate_marks, repeat=i):
                if should_end_mark is not None and comb[-1] != should_end_mark:
                    # cut down search space
                    continue
                combinations.append("".join(comb))
        return combinations

    def complete_json_object(self, json_str: str, completion_str: str) -> Any:
        """
        Complete a JSON object string by appending a completion string.

        Parameters
        ----------
        json_str : str
            The original JSON object string.
        completion_str : str
            The completion string to append.

        Returns
        -------
        Any
            The completed JSON object as a Python object.

        Raises
        ------
        ValueError
            If the JSON object string cannot be fixed.

        Notes:
        ------
        - This function appends the `completion_str` to the end of `json_str` until a valid JSON object can be obtained. If `json_str` is an invalid JSON object string, the function will remove one character from the end of `json_str` and try again until it either finds a valid JSON object string or until there are no more characters left to remove.

        Examples
        --------
        >>> complete_json_object('{"a": 1, "b": 2', '}')
        {'a': 1, 'b': 2}

        >>> complete_json_object('{"a": 1, "b": 2}}}}}}}}}}}}}}}}}}}', '')
        {'a': 1, 'b': 2}

        >>> complete_json_object('{"a": 1, "b": 2', '')
        Traceback (most recent call last):
            ...
        ValueError: Couldn't fix JSON
        """
        while True:
            if not json_str:
                raise ValueError("Couldn't fix JSON")
            try:
                complete_json_str = json_str + completion_str
                python_obj = eval(complete_json_str)
            except Exception:
                json_str = json_str[:-1]
                continue
            return python_obj

    def get_possible_completions(
        self, json_str: str, json_depth_limit: int = 5
    ) -> Union[Dict[str, Any], List[Any]]:
        """
        Returns a list of possible completions for a JSON object string.

        Parameters
        ----------
        json_str : str
            The JSON object string
        json_depth_limit : int, optional
            The maximum length of the completion strings to try (default is 5)

        Returns
        -------
        Union[Dict[str, Any], List[Any]]
            If the completion strings are objects, returns a dictionary with 'completion' and 'suggestions' keys.
            If the completion strings are arrays, returns a list of suggested completions.
        """
        candidate_marks = ["}", "]"]
        if "[" not in json_str:
            candidate_marks.remove("]")
        if "{" not in json_str:
            candidate_marks.remove("}")

        # specify the mark should end with
        should_end_mark = "]" if json_str.strip()[0] == "[" else "}"
        completions = []
        for completion_str in self.get_combinations(
            candidate_marks, json_depth_limit, should_end_mark=should_end_mark
        ):
            try:
                completed_obj = self.complete_json_object(json_str, completion_str)
                completions.append(completed_obj)
            except Exception:
                pass
        return self.find_max_length(completions)

    def fit(self, json_str: str, json_depth_limit: int = 5) -> Dict[str, Any]:
        """
        Tries to parse the input JSON string and complete it if it is incomplete.

        Parameters
        ----------
        json_str : str
            The input JSON string
        json_depth_limit : int, optional
            The maximum length of the completion strings to try (default is 5)

        Returns
        -------
        Dict[str, Any]
            A dictionary with 'status' and 'data' keys. If the status is 'completed', the 'data'
            key will contain the completed object and an empty list of suggestions. If the status
            is 'failed', the 'data' key will contain an error message string. If the status is
            'incomplete', the 'data' key will contain a list of possible completions.
        """
        try:
            output = eval(json_str)
            return {
                "status": "completed",
                "object_type": type(output),
                "data": {"completion": output, "suggestions": []},
            }
        except Exception:
            # remove tail braces or brackets to speed up searching.
            json_str = re.sub(r"[\[\]\{\}\s]+$", "", json_str)
            try:
                output = self.get_possible_completions(
                    json_str, json_depth_limit=json_depth_limit
                )
                return {
                    "status": "completed",
                    "object_type": type(output["completion"]),
                    "data": output,
                }
            except Exception as e:
                return {
                    "status": "failed",
                    "object_type": None,
                    "data": {"error_message": str(e)},
                }

    def find_max_length(self, data_list: List[Any]) -> Dict[str, List[Any]]:
        """
        Returns a dictionary containing the element with the maximum length in the input list,
        as well as a list of all elements sorted by length in descending order.

        Parameters
        ----------
        data_list : list of any type
            A list of elements to be compared by length

        Returns
        -------
        dict
            A dictionary with keys 'completion' and 'suggestions'.

            The value of 'completion' key is the element with the maximum length in the input list.

            The value of 'suggestions' key is a list of all elements sorted by length in descending order.
        """
        # Create a dictionary where the keys are the indices of the elements in the input list
        # and the values are the lengths of the corresponding elements.
        length_dict = {i: len(str(element)) for i, element in enumerate(data_list)}

        # Sort the dictionary by value (length) in descending order.
        sorted_indices = sorted(length_dict.items(), key=itemgetter(1), reverse=True)

        # Create a new dictionary with the element with the maximum length as the 'completion' value
        # and a list of all elements sorted by length as the 'suggestions' value.
        output_dict = {
            "completion": data_list[sorted_indices[0][0]],
            "suggestions": [data_list[i] for i, _ in sorted_indices],
        }
        return output_dict

    def extract_complete_objects(self, string: str) -> List[Any]:
        """
        Extracts all complete Python objects from a string.

        Parameters
        ----------
        string : str
            The string to extract objects from.

        Returns
        -------
        List[Any]
            A list of all complete Python objects found in the string.
        """
        object_regex = r"(?<!\\)(\[[^][]*?(?<!\\)\]|\{[^{}]*\})"

        # The regular expression pattern matches any string starting with an opening brace or bracket,
        # followed by any number of non-brace and non-bracket characters, and ending with a closing brace
        # or bracket that is not preceded by an odd number of backslash escape characters.

        object_strings = []
        opening = {"{": 0, "[": 0}
        closing = {"}": "{", "]": "["}
        stack = []
        start = 0
        for match in re.finditer(object_regex, string):
            if len(stack) == 0:
                start = match.start()
            stack.append(match.group(1))
            if match.group(1)[-1] in closing:
                opening_bracket = closing[match.group(1)[-1]]
                opening[opening_bracket] += 1
                if opening[opening_bracket] == len(
                    [bracket for bracket in opening.values() if bracket != 0]
                ):
                    object_strings.append(string[start : match.end()])
                    stack = []
                    opening = {"{": 0, "[": 0}
                    closing = {"}": "{", "]": "["}
        if len(stack) > 0:
            print(f"Error: Incomplete object at end of string: {stack[-1]}")
        objects = []
        for object_string in object_strings:
            try:
                obj = ast.literal_eval(object_string)
                # Use ast.literal_eval() to safely evaluate the string as a Python object.
                objects.append(obj)
            except (ValueError, SyntaxError) as e:
                # If the string cannot be safely evaluated as a Python object, log an error and move on to the next object.
                print(f"Error evaluating object string '{object_string}': {str(e)}")
                pass

        return objects



__version__ = "0.1.4"
from .parser.parser import Parser
from .prompter.nlp_prompter import Prompter
from .prompter.prompt_cache import PromptCache
from .prompter.template_loader import TemplateLoader
from .prompter.conversation_logger import ConversationLogger
from .models.text2text.api.openai_models import OpenAI
from .models.text2text.api.anthropic import AnthropicModel
from .models.text2text.api.cohere import CohereModel
from .models.text2text.api.azure_openai import Azure
from .models.text2text.api.hub_model import HubModel
from .models.text2text.api.mock_model import MockModel
from .models.text2text.api.base_model import Model
from .utils.file_utils import *
from .utils.data_utils import *
from .utils.conversation_utils import *
from .pipelines import Pipeline



import json
import uuid
import datetime
from typing import Dict, Any


def get_conversation_schema(
    conversation_id: str, llm_name: str, **llm_metadata: Any
) -> Dict[str, Any]:
    """
    Constructs a conversation schema with the specified parameters.

    Args:
    - conversation_id: A string representing the unique identifier of the conversation.
    - llm_name: A string representing the name of the language model.
    - **llm_metadata: Optional additional metadata to associate with the language model.

    Returns:
    A dictionary representing the conversation schema.
    """
    # Remove any api_key from the kwargs to avoid potential security issues
    llm_metadata.pop("api_key", None)

    # Construct the conversation schema dictionary
    conversation_schema = {
        "conversation_id": conversation_id,
        "start_time": str(datetime.datetime.now().strftime("%Y_%m_%d:%H:%M:%S")),
        "llm": {"name": llm_name, "meta_data": llm_metadata},
        "participants": [
            {"name": "User", "is_bot": False},
            {"name": "Assistant", "is_bot": True},
        ],
        "messages": [],
    }

    return conversation_schema


def create_message(
    task: str,
    prompt: str,
    response: str,
    structured_response: Any,
    prompt_file: str,
    **template_metadata: Any
) -> Dict[str, Any]:
    """
    Creates a message dictionary with the specified parameters.

    Args:
    - task: A string representing the task the message is associated with.
    - prompt: A string representing the prompt that initiated the message.
    - response: A string representing the message response.
    - structured_response: A structured representation of the message response.
    - **template_metadata: Optional metadata to associate with the message.

    Returns:
    A dictionary representing the message.
    """
    # Get the current timestamp as a formatted string
    timestamp = str(datetime.datetime.now().strftime("%Y_%m_%d:%H:%M:%S"))
    prompt_id = str(uuid.uuid4())

    # Construct the message dictionary
    
    message = {
        "request_timestamp": timestamp,
        "prompt_id": prompt_id,
        "prompt_filename": prompt_file, 
        "processing_task": task,
        "template_metadata": template_metadata,
        "input_prompt": prompt,
        "response_text": response,
        "parsed_output": structured_response,
        }

    return message





import pytest
from promptify import HubModel
from typing import List, Dict
import requests


@pytest.fixture
def huggingface_complete():
    # set api key for testing -> https://huggingface.co/docs/hub/security-tokens
    huggingface_complete = HubModel(api_key = "", model= "https://api-inference.huggingface.co/models/mrm8488/t5-base-finetuned-common_gen", api_wait=1, api_retry=1)
    return huggingface_complete


def test_supported_models(huggingface_complete):
    models = huggingface_complete.supported_models()
    assert "mrm8488/t5-base-finetuned-common_gen" in models


def test_invalid_model(huggingface_complete):
    with pytest.raises(ValueError):
        huggingface_complete.set_model("invalid-model")


def test_get_description(huggingface_complete):
    description = huggingface_complete.get_description()
    assert isinstance(description, str)


def test_get_endpoint(huggingface_complete):
    endpoint = huggingface_complete.get_endpoint()
    assert isinstance(endpoint, str)


def test_execute_with_query(huggingface_complete):
    prompts = ["Hello, my name is", "The quick brown fox jumps over the"]
    results = huggingface_complete.model_output(huggingface_complete.execute_with_retry(prompts))
    assert isinstance(results, dict)
    assert "text" in results
    assert len(results["text"]) == 2
    assert isinstance(results["text"], list)

    for result in results["text"]:
        assert isinstance(result, List)
        for text in result:
            assert isinstance(text, str)
            assert text != ""



import json
import uuid
import datetime
from pathlib import Path
import hashlib
import os
import io


def read_json(json_file):
    """
    Reads JSON data from a file and returns a Python object.

    Args:
        json_file (str): The path to the JSON file to read.

    Returns:
        A Python object representing the JSON data.
    """
    with open(json_file) as f:
        try:
            data = json.load(f)
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Error decoding JSON data from file {json_file}: {str(e)}"
            )
    return data


def write_json(path, data, file_name):
    """
    Writes JSON data to a file.

    Args:
        path (str): The path to the directory where the file should be saved.
        data (Any): The data to write to the file. This can be any JSON-serializable object.
        file_name (str): The name of the file to write, without the '.json' extension.

    Raises:
        IOError: If there is a problem writing the file.
    """
    full_path = os.path.join(path, f"{file_name}.json")
    try:
        with open(full_path, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)
    except IOError as e:
        raise IOError(f"Error writing JSON file '{full_path}': {e.strerror}")


def calculate_hash(text: str, encoding: str = "utf-8") -> str:
    """
    Calculate the hash of a text using the specified encoding.

    Args:
        text: The text to calculate the hash for.
        encoding: The encoding to use for the text. Defaults to "utf-8".

    Returns:
        The hash of the text.
    """
    if not isinstance(text, str):
        raise TypeError("Expected a string for 'text' parameter.")

    hash_obj = hashlib.md5()
    hash_obj.update(text.encode(encoding))
    return hash_obj.hexdigest()


def setup_folder(folder_path: str, folder_name: str = None) -> str:
    """
    Creates a folder in the specified folder_path.

    Parameters
    ----------
    folder_path : str
        The path to the directory where the folder will be created.
    folder_name : str, optional
        The name of the folder. If None, a name will be generated using the
        current date and time and a UUID.

    Returns
    -------
    The path to the created folder.
    """

    if folder_name is None:
        current_date = datetime.datetime.now().strftime("%Y_%m_%d:%H:%M:%S")
        conversation_id = uuid.uuid4()
        folder_name = f"{current_date}_{conversation_id}"
        folder_name = calculate_hash(folder_name)

    folder_path = Path(folder_path)
    folder_path.mkdir(parents=True, exist_ok=True)

    folder = folder_path / folder_name
    folder.mkdir(parents=True, exist_ok=True)

    return str(folder), folder_name



import os
from glob import glob
from typing import List
from promptify.utils.file_utils import *
from jinja2 import Template, Environment, FileSystemLoader, meta


class TemplateLoader:
    """
    A class for loading and managing Jinja2 templates. It allows loading templates from files or strings,
    listing available templates, and getting template variables.
    """

    def __init__(self):
        """
        Initialize the TemplateLoader object and create an empty dictionary for loaded templates.
        """
        self.loaded_templates = {}

    def load_template(
        self, template: str, model_name: str, from_string: bool = False
    ):
        """
        Load a Jinja2 template either from a string or a file.

        Args:
            template (str): Template string or path to the template file.
            from_string (bool): Whether to load the template from a string. Defaults to False.

        Returns:
            dict: Loaded template data.
        """
        if template in self.loaded_templates:
            return self.loaded_templates[template]

        if from_string:
            template_instance = Template(template)
            template_data = {
                "template_name": "from_string",
                "template_dir": None,
                "environment": None,
                "template": template_instance,
            }
        else:
            template_data = self._load_template_from_path(template, model_name)

        self.loaded_templates[template] = template_data
        return self.loaded_templates[template]

    def _load_template_from_path(self, template: str, model_name: str) -> dict:
        """
        Load a Jinja2 template from the given path.

        Args:
            template (str): Path to the template file.

        Returns:
            dict: Loaded template data.
        """
        current_dir = os.path.dirname(os.path.realpath(__file__))
        current_dir, _ = os.path.split(current_dir)
        templates_dir = os.path.join(current_dir, "prompts", "text2text")
        all_folders = {
            f"{folder}.jinja": folder for folder in os.listdir(templates_dir)
        }

        if template in all_folders:
            meta_data = self._get_metadata(template, templates_dir, model_name)

            template_name = meta_data["metadata"]["file_name"]
            template_dir = meta_data["metadata"]["file_path"]
            environment = Environment(loader=FileSystemLoader(template_dir))
            template_instance = environment.get_template(template_name)

        else:
            self._verify_template_path(template)
            custom_template_dir, custom_template_name = os.path.split(template)

            template_name = custom_template_name
            template_dir = custom_template_dir
            environment = Environment(loader=FileSystemLoader(template_dir))
            template_instance = environment.get_template(custom_template_name)

        return {
            "template_name": template_name,
            "template_dir": template_dir,
            "environment": environment,
            "template": template_instance,
        }


    def search_model(self, data, model_name):

        all_models = []
        for sample in data:
            if model_name in sample['models']:
                return sample
            else:
                all_models.extend(sample['models'])
        raise ValueError(
                    f"Model not found. Please choose the model from : {all_models}"
                )

    def _get_metadata(self, template_name, template_path, model_name):

        template_name, _ = template_name.split(".jinja")
        metadata_files = glob(
            os.path.join(template_path, template_name, "metadata.json")
        )

        metadata = read_json(metadata_files[0])
        metadata = self.search_model(metadata, model_name)
        metadata["file_path"] = os.path.join(template_path, template_name)
    
        return {"metadata": metadata}

    def _verify_template_path(self, templates_path: str):
        """
        Verify the existence of the template file.

        Args:
            templates_path (str): Path to the template file.

        Raises:
            ValueError: If the template file does not exist.
        """
        if not os.path.isfile(templates_path):
            raise ValueError(f"Templates path {templates_path} does not exist")

    def list_templates(self, environment) -> List[str]:
        """
        List all templates in the specified environment.

            Args:
                environment (Environment): The Jinja2 environment to search for templates.

        Returns:
            List[str]: List of available template names.
        """
        return environment.list_templates()

    def get_template_variables(self, environment, template_name) -> List[str]:
        """
        Get a list of undeclared variables for the specified template.

        Args:
            environment (Environment): The Jinja2 environment of the template.
            template_name (str): The name of the template.

        Returns:
            List[str]: List of undeclared variables in the template.
        """
        template_source = environment.loader.get_source(environment, template_name)
        parsed_content = environment.parse(template_source)
        return list(meta.find_undeclared_variables(parsed_content))



from pathlib import Path
from tqdm import tqdm
from typing import Any, Dict, Optional
from promptify.prompter.conversation_logger import *
from promptify.utils.data_utils import *
from promptify.prompter.prompt_cache import PromptCache


class Pipeline:
    def __init__(self, prompter, model, structured_output=True, **kwargs):
        if not isinstance(prompter, list):
            prompter = [prompter]

        self.prompters = prompter
        self.model = model
        self.json_depth_limit: int = kwargs.get("json_depth_limit", 20)
        self.cache_prompt = kwargs.get("cache_prompt", True)
        self.cache_size = kwargs.get("cache_size", 200)
        self.prompt_cache = PromptCache(self.cache_size)
        self.conversation_path = kwargs.get("output_path", Path.cwd())
        self.structured_output = structured_output

        self.model_args_count = self.model.run.__code__.co_argcount
        self.model_variables = self.model.run.__code__.co_varnames[
            1 : self.model_args_count
        ]

        self.conversation_path = os.getcwd()
        self.model_dict = {
            key: value
            for key, value in model.__dict__.items()
            if is_string_or_digit(value)
        }
        self.logger = ConversationLogger(self.conversation_path, self.model_dict)

    def fit(self, text_input: str, **kwargs) -> Any:
        """
        Processes an input text through the pipeline: generates a prompt, gets a response from the model,
        caches the response, logs the conversation, and returns the output.
        """

        outputs_list = []
        for prompter in tqdm(self.prompters):
            try:
                template, variables_dict = prompter.generate(text_input, self.model.model, **kwargs)

            except ValueError as e:
                print(f"Error in generating prompt: {e}")
                return None

            if kwargs.get("verbose", False):
                print(template)

            output = self._get_output_from_cache_or_model(template)

            if output is None:
                return None

            if "jinja" in prompter.template:
                prompt_name = prompter.template
            else:
                prompt_name = "Unknown"

            if self.structured_output:
                message = create_message(
                    template,
                    variables_dict,
                    output["text"],
                    output["parsed"]["data"]["completion"],
                    prompt_name,
                )
            else:
                message = create_message(
                    template, variables_dict, output, None, prompt_name
                )

            self.logger.add_message(message)
            outputs_list.append(output)

        return outputs_list

    def _get_output_from_cache_or_model(self, template):
        output = None

        if self.cache_prompt:
            output = self.prompt_cache.get(template)

        if output is None:
            try:
                response = self.model.execute_with_retry(prompt=template)
            except Exception as e:
                print(f"Error in model execution: {e}")
                return None

            if self.structured_output:
                output = self.model.model_output(
                    response, json_depth_limit=self.json_depth_limit
                )
            else:
                output = response

            if self.cache_prompt:
                self.prompt_cache.add(template, output)

        return output



import pytest
from promptify import Model
from promptify import MockModel

@pytest.fixture
def mock_model():
    return MockModel("api_key", "mock_model", api_wait=5, api_retry=5)

def test_mock_model_is_instance_of_model(mock_model):
    assert isinstance(mock_model, Model)

def test_supported_models(mock_model):
    assert mock_model.supported_models() == ["mock_model"]

def test_init(mock_model):
    assert mock_model.api_key == "api_key"
    assert mock_model.model == "mock_model"

def test_get_description(mock_model):
    assert mock_model.get_description() == "Mock model for testing purposes"

def test_get_endpoint(mock_model):
    assert mock_model.get_endpoint() == "https://mock.endpoint/"

def test_get_parameters(mock_model):
    assert mock_model.get_parameters() == {"param": "value"}

def test_run(mock_model):
    prompts = ["prompt1", "prompt2"]
    responses = mock_model.run(prompts)
    assert responses == ["response", "response"]

def test_model_output(mock_model):
    response = "response"
    assert mock_model.model_output(response) == response

def test_execute_with_retry(mock_model):
    prompts = ["prompt1", "prompt2"]
    responses = mock_model.execute_with_retry(prompts)
    assert responses == ["response", "response"]



import json

def is_string_or_digit(obj):
    """
    Check if an object is a string or a digit (integer or float).

    Args:
        obj (any): The object to be checked.

    Returns:
        bool: True if the object is a string or a digit, False otherwise.

    Examples:
        >>> is_string_or_digit("hello")
        True
        >>> is_string_or_digit(123)
        True
        >>> is_string_or_digit(3.14)
        True
        >>> is_string_or_digit(True)
        False
    """
    return isinstance(obj, (str, int, float))



from typing import Dict, List, Optional, Tuple, Union
from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT
from promptify.parser.parser import Parser
from promptify.models.text2text.api.base_model import Model
import re

class AnthropicModel(Model):
    name = "Anthropic"
    description = "Anthropic API for text completion using Claude model"

    SUPPORTED_MODELS = set(
        [
            "claude-instant-1",
            "claude-2"
        ]
    )

    POSSIBLE_PREFIX = ['here are', 'here is', 'following are']

    def __init__(
        self,
        api_key: str,
        model: str = "claude-instant-1",
        temperature: float = 1,
        top_p: float = 0.7,
        top_k: int = 5,
        max_tokens_to_sample: int = 300,
        stop_sequences: Optional[Union[str, List[str]]] = ["\n\nHuman:"],
        api_wait=60,
        api_retry=6,
        json_depth_limit: int = 20
    ):
        super().__init__(api_key, model, api_wait, api_retry)

        self.temperature = temperature
        self.top_p = top_p
        self.top_k = top_k
        self.max_tokens_to_sample = max_tokens_to_sample
        self.json_depth_limit = json_depth_limit
        self.stop_sequences = stop_sequences
        self.set_key(api_key)
        self._verify_model()
        self.parameters = self.get_parameters()
        self._initialize_parser()

        

    def set_key(self, api_key: str):
        self._anthropic = Anthropic(api_key=api_key)

    def _verify_model(self):
        if self.model not in self.SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model: {self.model}")

    def set_model(self, model: str):
        self.model = model
        self._verify_model()

    def _initialize_parser(self):
        self.parser = Parser()

    def supported_models(self):
        return list(self.SUPPORTED_MODELS)

    def get_parameters(self):
        return {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "max_tokens_to_sample": self.max_tokens_to_sample,
            "stop_sequences": self.stop_sequences,
        }

    def get_description(self):
        return self.description

    def get_endpoint(self):
        return self.model

    def run(self, prompt: str):
        self.parameters["prompt"] = f"{HUMAN_PROMPT} {prompt} {AI_PROMPT}"
        response = self._anthropic.completions.create(
            model=self.model,
            **self.parameters,
        )
        return response

    def extract_string_json(self, text):
        pattern = r'\[\s*(\{.*?\})\s*\]'
        match = re.search(pattern, text, flags=re.S)
        if match:
            try:
                return match.group()
            except json.JSONDecodeError as e:
                raise f"Error decoding JSON: {e}"

    def model_output_raw(self, response: Dict) -> Dict:
        data = {}
        raw_response = response.completion.strip(" \n")
        
        string_json = []
        for prefix in self.POSSIBLE_PREFIX:
            if prefix in raw_response.lower():
                string_json.append(self.extract_string_json(str(raw_response)))
                break
        
        if string_json:
            data['text'] = str(string_json)
        else:
            data['text'] = str(raw_response)
        return data

    def model_output(self, response, json_depth_limit) -> Dict:
        data = self.model_output_raw(response)
        data["parsed"] = self.parser.fit(data["text"], json_depth_limit)
        return data



from promptify.utils.file_utils import *
from promptify.utils.conversation_utils import *
from promptify.utils.data_utils import *


class ConversationLogger:
    def __init__(self, conversation_path: str, llm_parameters: Dict):
        """Create a logger for a conversation.

        Args:
            conversation_path: The path to the folder where conversations will be stored.
            model_dict: A dictionary containing metadata about the model.
        """

        self.conversation_id = str(uuid.uuid4())
        self.storage_name = f"llm_responses/llm_session_{self.conversation_id}/"
        self.conversation_path = os.path.join(conversation_path, self.storage_name)
        Path(self.conversation_path).mkdir(parents=True, exist_ok=True)
        self.llm_parameters = llm_parameters
        self.model_dict = {
            key: value
            for key, value in self.llm_parameters.items()
            if is_string_or_digit(value)
        }

    def add_message(self, message: Dict[str, Any]):
        """Add a message to the conversation.
        Args:
            prompt: The prompt sent to the language model.
            response: The response generated by the language model.
            source: The source of the message ("user" or "model").
            metadata: Optional metadata about the message.
            **kwargs: Additional metadata to be added to the message.
        """

        self.conversation = get_conversation_schema(
            self.conversation_id, self.llm_parameters["model"], **self.llm_parameters
        )
        message_id = str(uuid.uuid4())
        self.conversation["messages"].append(message)
        write_json(self.conversation_path, self.conversation, message_id)

    def __repr__(self):
        return f"ConversationLogger(conversation_id={self.conversation_id}, conversation_path={self.conversation_path})"



from promptify.parser.parser import Parser
from promptify.models.text2text.api.base_model import Model
from typing import Dict, List, Optional, Tuple, Union
import cohere

class CohereModel(Model):
    name = "Cohere"
    description = "Cohere API for text completion using Command model"

    SUPPORTED_MODELS = set(
        [
            "command", 
            "command-light", 
            "command-medium-beta", 
            "command-nightly", 
            "command-xlarge-beta"
        ]
    )
    
    def __init__(
        self,
        api_key: str,
        model: str = "command",
        temperature: float = 0.9,
        top_k: int = 0,
        max_tokens: int = 200,
        stop_sequences: Optional[Union[str, List[str]]] = [],
        return_likelihoods: str = 'NONE',
        stream: bool = False,
        truncate: str = 'END',
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        top_p: float = 0.75,
        num_generations: int = 1,
        api_wait=60,
        api_retry=6,
        json_depth_limit: int = 20
    ):
        super().__init__(api_key, model, api_wait, api_retry)

        self.temperature = temperature
        self.top_k = top_k
        self.max_tokens = max_tokens
        self.stop_sequences = stop_sequences
        self.return_likelihoods = return_likelihoods
        self.json_depth_limit = json_depth_limit
        self.stream = stream
        self.truncate = truncate
        self.presence_penalty = presence_penalty
        self.frequency_penalty = frequency_penalty
        self.top_p = top_p
        self.num_generations = num_generations
        self.set_key(api_key)
        self._verify_model()
        self.parameters = self.get_parameters()
        self._initialize_parser()


    # https://dashboard.cohere.ai/api-keys

    def set_key(self, api_key: str):
        self._cohere = cohere.Client(api_key)

    def _verify_model(self):
        if self.model not in self.SUPPORTED_MODELS:
            raise ValueError(f"Unsupported model: {self.model}")
            
    def _initialize_parser(self):
        self.parser = Parser()

    def set_model(self, model: str):
        self.model = model
        self._verify_model()

    def supported_models(self):
        return list(self.SUPPORTED_MODELS)

    def get_parameters(self):
        return {
            "temperature": self.temperature,
            "k": self.top_k,
            "max_tokens": self.max_tokens,
            "stop_sequences": self.stop_sequences,
            "return_likelihoods": self.return_likelihoods,
            "stream": self.stream,
            "truncate": self.truncate,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
            "p": self.top_p,
            "num_generations": self.num_generations
        }

    def get_description(self):
        return self.description

    def get_endpoint(self):
        return self.model

    def run(self, prompt: str):
        self.parameters["prompt"] = prompt
        response = self._cohere.generate(
            model=self.model,
            **self.parameters,
        )
        return response

    def model_output_raw(self, response: Dict) -> Dict:
        data = {}
        raw_response = response.generations[0].text.strip(" \n")
        data['text'] = str(raw_response)
        return data
    
    def model_output(self, response, json_depth_limit) -> Dict:
        data = self.model_output_raw(response)
        data["parsed"] = self.parser.fit(data["text"], json_depth_limit)
        return data



from promptify import Prompter,OpenAI, Pipeline

sentence     =  "The patient is a 93-year-old female with a medical  				 
                history of chronic right hip pain, osteoporosis,					
                hypertension, depression, and chronic atrial						
                fibrillation admitted for evaluation and management				
                of severe nausea and vomiting and urinary tract				
                infection"

model        = OpenAI(api_key) # or `HubModel()` for Huggingface-based inference or 'Azure' etc
prompter     = Prompter('ner.jinja') # select a template or provide custom template
pipe         = Pipeline(prompter , model)

output = pipe.fit(text_input=sentence, domain="medical", labels=None)
print(output)



import os
import uuid
from glob import glob
import datetime
from pathlib import Path

from promptify.utils.data_utils import *
from promptify.prompter.template_loader import TemplateLoader

from typing import List, Dict, Any, Optional
from jinja2 import Environment, FileSystemLoader, meta, Template


class Prompter:

    """
    A class to generate and manage prompts.

    """

    def __init__(
        self,
        template,
        from_string = False,
        allowed_missing_variables: Optional[List[str]] = None,
        default_variable_values: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Initialize Prompter with default or user-specified settings.

        Parameters
        ----------
        template : str, optional
            A Jinja2 template to use for generating the prompt. Must be a valid file path.
        allowed_missing_variables : list of str, optional
            A list of variable names that are allowed to be missing from the template. Default is ['examples', 'description', 'output_format'].
        default_variable_values : dict of str: any, optional
            A dictionary mapping variable names to default values to be used in the template.
            If a variable is not found in the input dictionary or in the default values, it will be assumed to be required and an error will be raised. Default is an empty dictionary.
        """

        self.template = template
        self.template_loader = TemplateLoader()
        self.allowed_missing_variables = [
            "examples",
            "description",
            "output_format",
        ]
        self.allowed_missing_variables.extend(allowed_missing_variables or [])
        self.default_variable_values = default_variable_values or {}
        self.from_string = from_string


    def update_default_variable_values(self, new_defaults: Dict[str, Any]) -> None:
        self.default_variable_values.update(new_defaults)

    def generate(self, text_input, model_name, **kwargs) -> str:
        """
        Generates a prompt based on a template and input variables.

        Parameters
        ----------
        text_input : str
            The input text to use in the prompt.
        **kwargs : dict
            Additional variables to be used in the template.

        Returns
        -------
        str
            The generated prompt string.
        """

        loader = self.template_loader.load_template(
            self.template, model_name, self.from_string
        )

        kwargs["text_input"] = text_input

        if loader["environment"]:
            variables = self.template_loader.get_template_variables(
                loader["environment"], loader["template_name"]
            )
            variables_dict = {
                temp_variable_: kwargs.get(temp_variable_, None)
                for temp_variable_ in variables
            }

            variables_missing = [
                variable
                for variable in variables
                if variable not in kwargs
                and variable not in self.allowed_missing_variables
                and variable not in self.default_variable_values
            ]

            if variables_missing:
                raise ValueError(
                    f"Missing required variables in template {', '.join(variables_missing)}"
                )
        else:
            variables_dict = {"data": None}

        kwargs.update(self.default_variable_values)
        prompt = loader["template"].render(**kwargs).strip()

        if kwargs.get("verbose", False):
            print(prompt)

        return prompt, variables_dict



from functools import lru_cache, cached_property

class PromptCache:
    def __init__(self, cache_size: int = 200):
        self.cache_size = cache_size
        self._cache     = lru_cache(maxsize=cache_size)(self._raw_cache)()

    @cached_property
    def cache(self):
        return self._cache

    def get(self, key):
        return self.cache.get(key)

    def add(self, key, value):
        if key not in self.cache:
            self.cache[key] = value

    def clear(self):
        self.cache.clear()

    @staticmethod
    def _raw_cache():
        return {}



from typing import Dict, List, Optional, Tuple, Union
from huggingface_hub import model_info
from promptify.models.text2text.api.base_model import Model
from huggingface_hub.utils import (
    RepositoryNotFoundError,
    build_hf_headers,
    hf_raise_for_status,
)

from promptify import __version__
import requests


class HubModel(Model):

    name = "HuggingFace"
    description = "HuggingFace API for text completion using various models"

    def __init__(
        self,
        api_key: Optional[str] = None,
        model: str = "google/flan-t5-xl",
        wait_for_model: bool = True,
        use_cache: bool = True,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        temperature: float = 1.0,
        repetition_penalty: Optional[float] = None,
        max_new_tokens: Optional[int] = 20,
        max_time: Optional[float] = None,
        num_return_sequences: int = 1,
        do_sample: bool = False,
        api_wait= 60,
        api_retry= 6,
        json_depth_limit: int = 20,
    ):
        

        self.set_key(api_key)
        self.set_model(model)
        super().__init__(self.api_key, self.model, api_wait, api_retry)

        self.wait_for_model = wait_for_model
        self.use_cache = use_cache
        self.top_k = top_k
        self.top_p = top_p
        self.temperature = temperature
        self.repetition_penalty = repetition_penalty
        self.max_new_tokens = max_new_tokens
        self.max_time = max_time
        self.num_return_sequences = num_return_sequences
        self.do_sample = do_sample
        self.json_depth_limit = json_depth_limit

        self._verify_model()
        self.set_key(self.api_key)

    @classmethod
    def supported_models(cls) -> Dict[str, str]:

        response = requests.get(
            "https://huggingface.co/api/models?pipeline_tag=text2text-generation&sort=downloads&direction=-1"
        )

        hf_raise_for_status(response)
        return {
            model: f"check more details at https://huggingface.co/{model}"
            for model in sorted(item["id"] for item in response.json())
        }

    def _verify_model(self):

        if self.model not in self.supported_models():
            raise ValueError(f"Unsupported model: {self.model}")

    def set_key(self, api_key: Optional[str]):

        self.api_key = api_key
        self._headers = build_hf_headers(
            token=api_key, library_name="promptify", library_version=__version__
        )

    def set_model(self, model: str):

        self.model = model
        if self.model.startswith("https://") and "huggingface" in self.model:
            # User provided a URL pointing to a Inference Endpoint
            self._url = self.model
            self.model = self.model.split("models/")[-1]
        else:
            self._url = self.get_endpoint()

    def get_description(self) -> str:

        return self.supported_models()[self.model]

    def model_output(self, response: Dict, json_depth_limit = None) -> Dict:
        
        return [item["generated_text"] for item in response.json()]
    
    def model_output_raw(self, response):
      return response.text
    
    
    def get_endpoint(self) -> str:
        try:
            info = model_info(self.model, token=self.api_key)
        except RepositoryNotFoundError:
            raise ValueError(
                f"Model '{self.model}' does not exist on the Huggingface Hub. Please visit https://huggingface.co/models to"
                " find a suitable model."
            )

        if info.pipeline_tag not in ("text-generation", "text2text-generation"):
            raise ValueError(
                f"Cannot use model {self.model}. Pipeline is of type {info.pipeline_tag}. Expecting either"
                " 'text-generation' or 'text2text-generation'."
            )

        return f"https://api-inference.huggingface.co/pipeline/{info.pipeline_tag}/{self.model}"

    def get_parameters(
        self,
    ) -> Dict[str, Union[str, int, float, List[str], Dict[str, int]]]:

        return {
            "wait_for_model": self.wait_for_model,
            "use_cache": self.use_cache,
            "top_k": self.top_k,
            "top_p": self.top_p,
            "temperature": self.temperature,
            "repetition_penalty": self.repetition_penalty,
            "max_new_tokens": self.max_new_tokens,
            "max_time": self.max_time,
            "num_return_sequences": self.num_return_sequences,
            "do_sample": self.do_sample,
        }

    def run(self, prompts: List[str]) -> List[Optional[str]]:
        result = []
        for prompt in prompts:
            response = requests.post(
                self._url,
                headers=self._headers,
                json={
                    "inputs": prompt,
                    "options": {
                        "wait_for_model": self.wait_for_model,
                        "use_cache": self.use_cache,
                    },
                    "parameters": {
                        "top_k": self.top_k,
                        "top_p": self.top_p,
                        "temperature": self.temperature,
                        "repetition_penalty": self.repetition_penalty,
                        "max_new_tokens": self.max_new_tokens,
                        "max_time": self.max_time,
                        "num_return_sequences": self.num_return_sequences,
                        "do_sample": self.do_sample,
                    },
                },
            )
            hf_raise_for_status(response)
            result.append(response)
        return result



import pytest
from promptify import Parser


@pytest.fixture
def parser():
    parser = Parser()
    return parser


def test_is_valid_json(parser):
    assert parser.is_valid_json('{"name": "Alice", "age": 30}')
    assert parser.is_valid_json("[1, 2, 3, 4]")
    assert not parser.is_valid_json('{"name": "Bob", "age": }')
    assert not parser.is_valid_json("not a JSON string")


def test_get_combinations(parser):
    combinations = parser.get_combinations(["}", "]"], 3, should_end_mark="}")
    assert combinations == ["}", "}}", "]}"]


def test_escaped_(parser):
    case_1 = """[[{'T': 'ANATOMY', 'E': 'immune system'},{'T': 'DISEASE', 'E': 'Parkinson's disease'},{'T': 'other', 'E': 'person's health'}]]"""
    case_2 = """[[{"T": "ANATOMY", "E": "immune system"},{"T": "DISEASE", "E": "Parkinson"s disease"},{"T": "other", "E": "person"s health"}]]"""
    case_3 = """[[{'T': 'ANATOMY', 'E': 'immune system'}, {'T': 'DISEASE', 'E': 'Parkinson disease'}, {'T': 'other', 'E': 'person health'}]]"""

    result_1 = parser.escaped_(case_1)
    assert eval(result_1) == [
        [
            {"T": "ANATOMY", "E": "immune system"},
            {"T": "DISEASE", "E": 'Parkinson"s disease'},
            {"T": "other", "E": 'person"s health'},
        ]
    ]

    result_2 = parser.escaped_(case_2)
    assert eval(result_2) == [
        [
            {"T": "ANATOMY", "E": "immune system"},
            {"T": "DISEASE", "E": "Parkinson's disease"},
            {"T": "other", "E": "person's health"},
        ]
    ]

    result_3 = parser.escaped_(case_3)
    assert eval(result_3) == [
        [
            {"T": "ANATOMY", "E": "immune system"},
            {"T": "DISEASE", "E": "Parkinson disease"},
            {"T": "other", "E": "person health"},
        ]
    ]


def test_complete_json_object(parser):
    assert parser.complete_json_object('{"a": 1, "b": 2', "}") == {"a": 1, "b": 2}
    assert parser.complete_json_object("[1, 2, 3", "]") == [1, 2, 3]

    with pytest.raises(ValueError):
        parser.complete_json_object('{"a": 1, "b": 2', "")


def test_get_possible_completions(parser):
    completion = parser.get_possible_completions('{"a": 1, "b": 2')["completion"]
    assert completion == {"a": 1, "b": 2}
    completion = parser.get_possible_completions("[1, 2, 3")["completion"]
    assert completion == [1, 2, 3]


def test_fit(parser):
    result = parser.fit('{"name": "Alice", "age":30}')

    case_1 = """[{'a' : 1, 'b' : 2}, {'a' : 1'"""
    assert parser.fit(case_1, 10)["data"]["completion"] == [{"a": 1, "b": 2}, {"a": 1}]

    case_2 = "[[1, 2, 3], [11, 12, 21]"
    assert parser.fit(case_2, 10)["data"]["completion"] == [[1, 2, 3], [11, 12, 21]]

    case_3 = """[{'a': [{'a': 1, 'b': 2}, {'a': 12, 'b': 23}], 'b': [{'a':"""
    assert parser.fit(case_3, 10)["data"]["completion"] == [
        {"a": [{"a": 1, "b": 2}, {"a": 12, "b": 23}], "b": [{"a"}]}
    ]

    case_4 = "[[{'a': [1, 2, 3], 'b': {'c': 4}}, {'d': 5}], {'e': {'f': {'g': 6}}}]"
    assert parser.fit(case_4, 10)["data"]["completion"] == [
        [{"a": [1, 2, 3], "b": {"c": 4}}, {"d": 5}],
        {"e": {"f": {"g": 6}}},
    ]

    case_5 = "[[{'a': [1, 2, 3], 'b': {'c': 4}}, {'d': 5}], {'e': {'f': {'g': 6]"
    assert parser.fit(case_5, 10)["data"]["completion"] == [
        [{"a": [1, 2, 3], "b": {"c": 4}}, {"d": 5}],
        {"e": {"f": {"g": 6}}},
    ]

    case_6 = "[{'a': 1}, {'b': 2"
    assert parser.fit(case_6, 10)["data"]["completion"] == [{"a": 1}, {"b": 2}]

    case_7 = """{"person": {"name": "Alice", "age": 30, "hobbies": ["reading", "running", {"favorite_movies": ["Inception", "The Matrix"]}, {"favorite_songs": ["Imagine", "Let it Be"]}"""
    assert parser.fit(case_7, 10)["data"]["completion"] == {
        "person": {
            "name": "Alice",
            "age": 30,
            "hobbies": [
                "reading",
                "running",
                {"favorite_movies": ["Inception", "The Matrix"]},
                {"favorite_songs": ["Imagine", "Let it Be"]},
            ],
        }
    }

    case_8 = """{"name": "Bob", "age": 25, "is_student": False, "scores": [85, 90, 78], "contact_info": {"email": "bob@example.com", "phone": "123-456-7890"}, "courses": [{"course_id": 101, "course_name": "Mathematics"}, {"course_id": 102, "course_name": "Physics"}"""
    assert parser.fit(case_8, 10)["data"]["completion"] == {
        "name": "Bob",
        "age": 25,
        "is_student": False,
        "scores": [85, 90, 78],
        "contact_info": {"email": "bob@example.com", "phone": "123-456-7890"},
        "courses": [
            {"course_id": 101, "course_name": "Mathematics"},
            {"course_id": 102, "course_name": "Physics"},
        ],
    }

    case_9 = """[1, 2, {"name": "Alice", "age": 30}, 4, [5, 6], {"a": 1, "b": [{"c": 2, "d": 3}], "e": 4}"""
    assert parser.fit(case_9, 10)["data"]["completion"] == [
        1,
        2,
        {"name": "Alice", "age": 30},
        4,
        [5, 6],
        {"a": 1, "b": [{"c": 2, "d": 3}], "e": 4},
    ]

    case_10 = """{"name": "Alice", "age": 30, "hobbies": ["reading", "running", {"favorite_movies": ["Inception", "The Matrix"]},"""
    assert parser.fit(case_10, 10)["data"]["completion"] == {
        "name": "Alice",
        "age": 30,
        "hobbies": [
            "reading",
            "running",
            {"favorite_movies": ["Inception", "The Matrix"]},
        ],
    }
